# -*- coding: utf-8 -*-
"""Tile functions.

This module contains functions used when working tiles.

"""
from geopandas import GeoDataFrame
from .utils import mask_to_shapely


def merge_tile_masks(tile_list: list, buffer: int = 1) -> GeoDataFrame:
    """
    Merge the tile masks into a single mask for a large image.

    Args:
        tile_list (list): List of three-lenght tuples containing:
            (1) fp or array for mask, (2) x-coordinate for tile, (3)
            y-coordinate for tile.
        buffer (int): Buffer to apply to the polygons before dissolving.

    Returns:
        GeoDataFrame: GeoDataFrame with the merged mask.

    """
    polygons_and_labels = []

    # Process each tile.
    for tile_info in tile_list:
        tile, x, y = tile_info

        # Process the mask by converting it to polygons.
        polygons_and_labels.extend(
            mask_to_shapely(tile, x_offset=x, y_offset=y)
        )

    # Convert polygons and labels into a GeoDataFrame.
    gdf = GeoDataFrame(polygons_and_labels, columns=["geometry", "label"])

    # Apply a buffer to make edges touch.
    gdf["geometry"] = gdf["geometry"].buffer(buffer)

    # Dissolve the dataframe by the label.
    gdf = gdf.dissolve(by="label", as_index=False)

    # Remove the buffer.
    gdf["geometry"] = gdf["geometry"].buffer(-buffer)

    return gdf



# Shadow imports.
from .imread import imread
from .imwrite import imwrite

# Make modules available.
from . import girder_utils
from . import dash
from . import ml
from . import mongo_utils
from . import tiling
from . import utils
from . import tile_utils

# Modules that should be available.
__all__ = [
    "girder_utils",
    "dash",
    "ml",
    "mongo_utils",
    "tiling",
    "utils",
    "tile_utils",
]



def format_ann_docs_for_paperjs(
    geojson_ann_docs: list[dict], return_as_input_to_paper: bool = False
) -> list[dict]:
    """Format DSA annotations in the geojson format into a format that paperdragon
    can read and show as polygons.

    Args:
        geojson_ann_docs (list[dict]): A list of geojson annotation documents.
        return_as_input_to_paper (bool): Whether to return the formatted documents as
            input to paper. Defaults to False.

    Returns:
        list[dict]: A list of formatted annotation documents for paperjs. Note that
            it will return an empty list if the documents could not be formatted.

    """
    # NOTE: hard-coded to only do the first doc.
    formatted_docs = []

    # Loop through each annotation document.
    for ann_doc in geojson_ann_docs:
        # Grab the features of this document, contains the polygon info.
        features_to_include = []

        for feature in ann_doc["features"]:
            geometry = feature["geometry"]
            if geometry["type"] == "Polygon":
                properties = feature["properties"]
                properties["rescale"] = {"strokeWidth": properties["lineWidth"]}
                properties["strokeColor"] = properties["lineColor"]
                properties["source"] = "dsa"

                # Paperjs only supports multipolygons.
                geometry["type"] = "MultiPolygon"

                coordinates = geometry["coordinates"]

                adjusted_coordinates = [
                    [[[int(coord) for coord in point[:2]] for point in coordinates[0]]]
                ]

                geometry["coordinates"] = adjusted_coordinates

                features_to_include.append(feature)

        if len(features_to_include):
            formatted_docs.append(features_to_include)

    if return_as_input_to_paper:
        return get_input_to_paper_dict(formatted_docs)

    return formatted_docs


def get_input_to_paper_dict(formatted_docs: list[dict] | None = None) -> dict:
    """Get the input to paper dictionary.

    Args:
        formatted_docs (list[dict]): A list of formatted annotation documents for paperjs.

    Returns:
        dict: The input to paper dictionary.

    """
    input_to_paper = [{"type": "clearItems"}]

    if formatted_docs is not None and len(formatted_docs):
        for doc in formatted_docs:
            input_to_paper.append(
                {
                    "type": "drawItems",
                    "itemList": doc,
                }
            )

    return {"actions": input_to_paper}



import numpy as np
from . import imread
import cv2 as cv

from shapely.geometry import Polygon
from shapely.affinity import translate


def mask_to_shapely(
    mask: str | np.ndarray,
    x_offset: int = 0,
    y_offset: int = 0,
    background_label: int = 0,
    min_area: int = 0,
) -> list[Polygon, int]:
    """
    Extract contours from a label mask and convert them into shapely
    polygons.

    Args:
        mask (str | np.ndarray): Path to the mask image or the mask.
        x_offset (int): Offset to add to x coordinates of polygons.
        y_offset (int): Offset to add to y coordinates of polygons.
        background_label (int): Label value of the background class,
            which is ignored.

    Returns:
        list[Polygon, int]: List of polygons and their corresponding
            labels.

    """
    if isinstance(mask, str):
        mask = imread(mask, grayscale=True)

    # Find unique labels (excluding background 0)
    labels = [label for label in np.unique(mask) if label != background_label]

    polygons = []  # Track all polygons.

    # Loop through unique label index.
    for label in labels:
        # Filter to mask for this label.
        label_mask = (mask == label).astype(np.uint8)

        # Find contours.
        contours, hierarchy = cv.findContours(
            label_mask, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE
        )

        # Process the contours.
        polygons_dict = {}

        for idx, (contour, h) in enumerate(zip(contours, hierarchy[0])):
            if len(contour) > 3:
                if idx not in polygons_dict:
                    polygons_dict[idx] = {"holes": []}

                if h[3] == -1:
                    polygons_dict[idx]["polygon"] = contour.reshape(-1, 2)
                else:
                    polygons_dict[h[3]]["holes"].append(contour.reshape(-1, 2))

        # Now that we know the polygon and the holes, create a Polygon object for each.
        for data in polygons_dict.values():
            if "polygon" in data:
                polygon = Polygon(data["polygon"], holes=data["holes"])

                # Shift the polygon by the offset.
                polygon = translate(polygon, xoff=x_offset, yoff=y_offset)

                # Skip small polygons.
                if polygon.area >= min_area:
                    polygons.append([polygon, label])

    return polygons



from girder_client import GirderClient
import pandas as pd
from pathlib import Path

from ...girder_utils import get_region, get_item_large_image_metadata
from ... import imwrite


def get_pearce_roi_images(
    gc: GirderClient,
    item_id: str,
    class_map: dict[str, int],
    save_dir: str = ".",
    mag: float | None = None,
) -> pd.DataFrame:
    """Download the smallestl bounding box around an ROI annotations and its
    accompanying object detection annotations. The format used to determine ROIs
    and boxes is based on the convention used by Dr. Thomas Pearce in his
    OpenSeadragon applications. Labels are saved as text files in the YOLO format.

    Args:
        gc (girder_client.GirderClient): The Girder client to use.
        item_id (str): The item ID of the WSI.
        class_map (dict[str, int]): A dictionary mapping class names to integers.
        save_dir (str, optional): The directory to save the images and labels.
            Defaults to ".".
        mag (float | None, optional): The magnification to download the ROI at.
            If None, the scan magnification is used. Defaults to None.

    Returns:
        pandas.DataFrame: A DataFrame containing metadata for each ROI

    """
    # Create directories to save images and labels.
    img_dir = Path(save_dir) / "images"
    img_dir.mkdir(parents=True, exist_ok=True)

    label_dir = Path(save_dir) / "labels"
    label_dir.mkdir(parents=True, exist_ok=True)

    # Get all the annotation documents with name starting with "ROI".
    ann_docs_ids = []

    for ann_doc in gc.get(
        f"annotation?itemId={item_id}&limit=0&offset=0&sort=lowerName&sortdir=1"
    ):
        if ann_doc.get("annotation", {}).get("name", "").startswith("ROI"):
            ann_docs_ids.append(ann_doc["_id"])

    if not len(ann_docs_ids):
        # No annotation documents, return None.
        return

    # Track metadata for each ROI.
    roi_metadata = []

    # Loop through each document.
    for ann_doc_id in ann_docs_ids:
        # Get the full annotation document.
        ann_doc = gc.get(f"annotation/{ann_doc_id}")

        # Separate the ROI element and the other elements.
        roi_el = None
        other_els = []

        for element in ann_doc.get("annotation", {}).get("elements", []):
            if element.get("user", {}).get("role") == "ROI":
                roi_el = element
            else:
                other_els.append(element)

        if roi_el is None:
            raise Exception("No ROI element found in annotation document.")

        # Download the ROI image.
        xc, yc = roi_el["center"][:2]
        roi_w, roi_h = int(roi_el["width"]), int(roi_el["height"])

        # Get the top left coordinate of the ROI.
        roi_left = int(xc - roi_w / 2)
        roi_top = int(yc - roi_h / 2)

        # Get scale factor to go from scan magnification to mag.
        scan_mag = get_item_large_image_metadata(gc, item_id)["magnification"]

        if mag is None:
            mag = scan_mag

        roi_img = get_region(
            gc, item_id, left=roi_left, top=roi_top, width=roi_w, height=roi_h, mag=mag
        )

        fn = f"itemId-{item_id}-x{roi_left}y{roi_top}x{roi_left+roi_w}y{roi_top+roi_h}"

        img_fp = img_dir / f"{fn}.png"
        imwrite(img_fp, roi_img)

        # Convert the boxes to YOLO format.
        labels = ""

        for element in other_els:
            el_class = element["user"]["class"]

            if el_class not in class_map:
                continue

            xc, yc = element["center"][:2]
            xc, yc = int(xc), int(yc)

            # Subtract the left and top of the ROI to get the correct coordinates.
            xc = xc - roi_left
            yc = yc - roi_top

            # Normalize by the ROI width and height.
            xc /= roi_w
            yc /= roi_h

            w = element["width"] / roi_w
            h = element["height"] / roi_h

            labels += f"{class_map[el_class]} {xc:.4f} {yc:.4f} {w:.4f} {h:.4f}\n"

        # Save the label to a text file.
        label_fp = label_dir / f"{fn}.txt"

        with open(label_fp, "w") as fh:
            fh.write(labels.strip())

        # Append all the metadata to the list.
        roi_metadata.append(
            [item_id, img_fp, roi_left, roi_top, roi_w, roi_h, mag, scan_mag]
        )

    return pd.DataFrame(
        roi_metadata, columns=["item_id", "fp", "x", "y", "w", "h", "mag", "scan_mag"]
    )



import pymongo
from os import getenv
from gridfs import GridFS
import numpy as np
from io import BytesIO
from bson.objectid import ObjectId
from PIL import Image
from typing import Union


def get_mongo_client(
    protocol: str | None = None,
    host: str | None = None,
    username: str | None = None,
    password: str | None = None,
    port: int | None = None,
):
    """Get a mongo client by providing the key parts of the URL. By default
    when the parameters are not passed, it sets them to default values first
    by looking at key environment variables. If the environment variables are
    not present it sets them to default values.

    Args:
        protocol (str, optional): The protocol of the mongo server. If None it
            look for the environment variable "MONGO_PROTOCOL". If this is
            missing it will default to "mongodb".
        host (str, optional): The host name of the mongo server. If None it will
            look for the environment variable "MONGO_HOST_NAME". If this is
            missing it will default to "mongodb".
        username (str, optional): The username to connect to the mongo server.
            If None it will look for the environment variable
            "MONGO_INITDB_ROOT_USERNAME". If this is missing it will default to
            "docker".
        password (str, optional): The password to connect to the mongo server.
            If None it will look for the environment variable
            "MONGO_INITDB_ROOT_PASSWORD". If this is missing it will default to
            "docker".
        port (int, optional): The port number of the mongo server. If None it
            will look for the environment variable "MONGO_HOST_PORT". If this is
            missing it will default to 27017.

    Returns:
        The mongo client.

    """
    if protocol is None:
        protocol = getenv("MONGO_PROTOCOL", "mongodb")
    if username is None:
        username = getenv("MONGO_INITDB_ROOT_USERNAME", "docker")
    if password is None:
        password = getenv("MONGO_INITDB_ROOT_PASSWORD", "docker")
    if host is None:
        host = getenv("MONGO_HOST_NAME", "mongodb")
    if port is None:
        port = int(getenv("MONGO_HOST_PORT", 27017))

    return pymongo.MongoClient(f"{protocol}://{username}:{password}@{host}:{port}")


def chunks(lst: list, n=500):
    """Helper function for traversting through a list in chunks.

    Args:
        lst (list): The list to traverse.
        n (int): The size of the chunks.

    Returns:
        generator: A generator of the list in chunks.

    """
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def add_many_to_collection(
    mongo_collection,
    items: dict | list,
    key: str = "_id",
) -> dict[str | dict]:
    """Add items to a mongo collection. For this project we always add items with a unique
    user key.

    Args:
        mongo_collection: The mongo collection to add to.
        items (dict | list): The items to add to the collection.
        key (str): The key to use to identify the items.

    Returns:
        dict[str | dict]: The items added to the collection.

    """
    # If the items are a list, then use the "key" parameter to nest it as a dict.
    if isinstance(items, list):
        items = {item[key]: item for item in items}

    operations = []

    for _id, item in items.items():
        operations.append(pymongo.UpdateOne({"_id": _id}, {"$set": item}, upsert=True))

    for chunk in chunks(operations):
        _ = mongo_collection.bulk_write(chunk)

    return items


def get_img_from_db(mongo_db, img_id: str) -> Union[np.ndarray, None]:
    """Get an image from the database by its location id.

    Args:
        mongo_db: The mongo database.
        img_loc (str): The image id in mongo.

    Returns:
        np.array: The image, if None then the image was not found.

    """
    fs = GridFS(mongo_db)

    grid_out = fs.get(ObjectId(img_id))

    if grid_out:
        # Read the byte data from the GridOut object
        byte_data = grid_out.read()

        # Create a BytesIO object from the byte data
        byte_io = BytesIO(byte_data)

        # Open the image from the BytesIO object
        src_img = np.array(Image.open(byte_io))

        return src_img


def add_img_to_db(
    mongo_db,
    img: Image.Image | np.ndarray,
) -> str:
    """Add an image to the database.

    Args:
        mongo_db: The mongo database.
        img (np.ndarray or PIL.Image.Image): The image.

    Returns:
        str: The image id in the database.

    """
    fs = GridFS(mongo_db)

    # Convert image to PIL image if needed.
    if isinstance(img, np.ndarray):
        img = Image.fromarray(img)

    # Create a BytesIO object and save the image to it
    byte_io = BytesIO()
    img.save(byte_io, "PNG")

    # Get the byte data from the BytesIO object
    byte_io.seek(0)
    byte_data = byte_io.read()

    # Save the byte data to MongoDB using GridFS
    file_id = fs.put(byte_data)

    return file_id



# -*- coding: utf-8 -*-
"""Metric functions.

This module contains functions used to evaluate the performance of 
machine learning models.

"""
import torch
from torch import nn
import numpy as np


def binary_dice_coefficient(eval_pred: tuple[np.ndarray, np.ndarray]) -> dict:
    """Calculate the dice coefficient for binary masks. Developed for use
    in compute_metrics parameters of transformers Trainer class, for
    SegFormer senantic segmentation model of two classes.

    Args:
        eval_pred (tuple[np.ndarray, np.ndarray]): A tuple containing the
            predicted and true masks.

    Returns:
        dict: A dictionary containing the dice coefficient.

    """
    with torch.no_grad():
        logits, labels = eval_pred

        # Convert to tensors.
        logits = torch.from_numpy(logits).cpu()

        # Resize the logits to be the same shape as the labels.
        logits = nn.functional.interpolate(
            logits,
            size=labels.shape[-2:],
            mode="bilinear",
            align_corners=False,
        ).argmax(dim=1)

        preds = logits.detach().numpy()

        # Convert the masks to binary.
        preds = preds > 0
        labels = labels > 0

        # Calculate dice.
        intersection = np.sum(preds * labels)
        union = np.sum(preds) + np.sum(labels)

        if union == 0:
            dice = 1.0  # Both masks are empty
        else:
            dice = 2.0 * intersection / union

        return {"dice_coefficient": dice}


def mean_iou(eval_pred: tuple[np.ndarray, np.ndarray]) -> dict:
    """Compute the mean intersection over union (IoU) for semantic
    segmentation. Also returns the IoU for each class.

    Args:
        eval_pred (tuple[numpy.ndarray, numpy.ndarray]): Tuple of
            logits and labels.

    Returns:
        dict: Dictionary of metrics.

    """
    with torch.no_grad():
        logits, labels = eval_pred

        # Convert logits to tensor.
        logits_tensor = torch.from_numpy(logits).cpu()
        labels_tensor = torch.from_numpy(labels).cpu()

        # From logits get the number of classes in the dataset.
        num_classes = logits.shape[1]

        # Scale the logits back to the shape of the labels.
        logits_tensor = nn.functional.interpolate(
            logits_tensor,
            size=labels.shape[-2:],
            mode="bilinear",
            align_corners=False,
        )

        # Turn the logits to predictions.
        pred = logits_tensor.argmax(dim=1)

        # Calculate the IoU for each class.
        ious = []

        metrics = {}

        for cls in range(num_classes):
            pred_mask = pred == cls
            labels_mask = labels_tensor == cls

            intersection = (
                torch.logical_and(pred_mask, labels_mask).sum().item()
            )
            union = torch.logical_or(pred_mask, labels_mask).sum().item()

            if union == 0:
                iou = float("nan")  # or 1.0
            else:
                iou = intersection / union

            metrics[f"label {cls} IoU"] = iou

            ious.append(iou)

        mean_iou = np.nanmean(ious)  # Use NumPy for NaN handling
        metrics["mean_iou"] = mean_iou

        return metrics



import numpy as np
from pathlib import Path
import cv2 as cv
import shutil

import transformers
from transformers import SegformerImageProcessor

import torch
import torch.nn as nn

from ... import imread
from ...tiling import tile_image


def inference(
    img: np.ndarray,
    model: transformers.models.segformer.modeling_segformer.SegformerForSemanticSegmentation,
    device: str | None = None,
    tile_size: int = 512,
    temp_dir: str = ".temp",
    batch_size: int = 8,
) -> tuple[np.ndarray, list[np.ndarray]]:
    """Inference on an image using SegFormer model for semantic segmentation.

    Args:
        img (numpy.ndarray): Image to segment.
        model (transformers.models.segformer.modeling_segformer.SegformerForSemanticSegmentation):
            Pre-trained model.
        device (str | None): Device to use. Defaults to None, in which case
            it is inferred based on devices available.
        tile_size (int): Size of the tiles. Defaults to 512.
        temp_dir (str): Temporary directory to save tiles. Defaults to ".temp".
        batch_size (int): Batch size. Defaults to 8.

    Returns:
        tuple[numpy.ndarray, list[numpy.ndarray]]: The predicted label mask and
            the contours of the mask.

    """
    # Create location to save tile images.
    tile_dir = Path(temp_dir)
    tile_dir.mkdir(exist_ok=True, parents=True)

    # Tile the image.
    tiles_df = tile_image(
        img,
        str(tile_dir),
        tile_size=tile_size,
    )

    # Create instance of the image processor.
    img_processor = SegformerImageProcessor()

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"

    print("Using device:", device)

    # Set model to eval mode.
    model.to(device)  # to device
    model.eval()

    # Create an empty mask the size of the image.
    h, w = img.shape[:2]
    predicted_mask = np.zeros((h + tile_size, w + tile_size), dtype=np.uint8)

    # Loop through tile images in batch size.
    for i in range(0, len(tiles_df), batch_size):
        # Get batch of tile images.
        batch = tiles_df.iloc[i : i + batch_size]

        # Read all the images in the batch.
        images = [imread(fp) for fp in batch["fp"]]

        # Process all the images.
        inputs = img_processor(
            images=images,
            return_tensors="pt",
        )

        # Send to the same device as model.
        inputs = inputs.to(device)

        with torch.no_grad():
            # Sent inputs through the model.
            outputs = model(**inputs)

            # Get out the logits.
            logits = outputs.logits

            # Reshape the logits to the original image size.
            upsampled_logits = nn.functional.interpolate(
                logits,
                size=(tile_size, tile_size),  # (height, width)
                mode="bilinear",
                align_corners=False,
            )

            # Take the max value of the class dimension.
            pred_seg = upsampled_logits.argmax(dim=1)

            # Place tile predictions into the overall predicted mask.
            for j in range(len(batch)):
                r = batch.iloc[j]
                x, y = r["x"], r["y"]
                predicted_mask[y : y + tile_size, x : x + tile_size] = (
                    pred_seg[j].cpu().numpy()
                ).astype(np.uint8)

    # Remove the temp directory.
    shutil.rmtree(temp_dir)

    # Remove edges.
    predicted_mask = predicted_mask[:h, :w]

    # Extract the contours of the mask.
    contours = cv.findContours(
        predicted_mask, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE
    )[0]

    # Return the mask.
    return predicted_mask, contours



import large_image
import cv2 as cv
import numpy as np
from pathlib import Path
from ...imwrite import imwrite
from multiprocessing import Pool
from tqdm import tqdm


def _process_tile(
    ts,
    prepend_name,
    x,
    y,
    scan_tile_size,
    tile_size,
    mag,
    contours,
    sf,
    img_dir,
    mask_dir,
):
    fn = f"{prepend_name}x{x}y{y}.png"

    # Get the tile.
    tile_img = ts.getRegion(
        region={
            "left": x,
            "top": y,
            "right": x + scan_tile_size,
            "bottom": y + scan_tile_size,
        },
        format=large_image.constants.TILE_FORMAT_NUMPY,
        scale={"magnification": mag},
    )[0][:, :, :3]

    if tile_img.shape[:2] != (tile_size, tile_size):
        # Pad the image with zeroes to make it the desired size.
        tile_img = cv.copyMakeBorder(
            tile_img,
            0,
            tile_size - tile_img.shape[0],
            0,
            tile_size - tile_img.shape[1],
            cv.BORDER_CONSTANT,
            value=0,
        )

    # Create a blank tile mask.
    tile_mask = np.zeros(tile_img.shape[:2], dtype=np.uint8)

    # Loop through contours to draw on tile mask.
    for idx, contour_dict in contours.items():
        # Create a mask for the specific label.
        label_mask = np.zeros(tile_mask.shape, dtype=np.uint8)

        points = contour_dict["points"].copy()
        holes = contour_dict["holes"].copy()

        # For points and hole:
        # (1) Shift by top left of the current tile.
        # (2) Scale by be desired magnification.
        points = [((point - [x, y]) * sf).astype(int) for point in points]

        corrected_holes = []

        for hole in holes:
            if len(hole):
                corrected_holes.append(((hole - [x, y]) * sf).astype(int))

        # Use points to draw contours as positive.
        label_mask = cv.drawContours(label_mask, points, -1, 1, cv.FILLED)

        # Draw in the holes as background class zero.
        label_mask = cv.drawContours(label_mask, holes, -1, 0, cv.FILLED)

        # Apply the label to tile mask where label mask is positive.
        tile_mask[label_mask == 1] = idx

    # Save the image and mask.
    img_fp = img_dir / fn
    imwrite(img_fp, tile_img)
    imwrite(mask_dir / fn, tile_mask, grayscale=True)

    return str(img_fp), x, y


def semantic_seg_tile_wsi_with_dsa_annotations(
    wsi_fp: str,
    annotation_docs: list[dict],
    label2idx: dict,
    save_dir: str,
    tile_size: int,
    mag: float | None = None,
    prepend_name: str = "",
    nproc: int = 1,
) -> list[str]:
    """Tile a WSI with semantic segmentation label masks created from
    DSA annotations.

    Args:
        wsi_fp (str): file path to WSI.
        annotation_docs (list[dict]): list of annotation documents.
        label2idx (dict): mapping of label names to integer indices.
        save_dir (str): directory to save the tiled images.
        tile_size (int): size of the tile images at desired magnificaiton.
        mag (float | None, optional): magnification level of the WSI.
            If None, the function will use the default magnification
            level. Defaults to None.
        prepend_name (str, optional): prepend name to the created tile
            images and masks. Defaults to "".
        nproc (int, optional): number of processes to use for tiling.

    Returns:
        list[str]: A list of tuples: (tile file path, x, y coordinates
            of tile).

    """
    # Format annotation documents for drawing on tile masks.
    contours = {idx: {"points": [], "holes": []} for idx in label2idx.values()}

    for ann_doc in annotation_docs:
        # Loop through the elements of the annotation document.
        elements = ann_doc["annotation"].get("elements", [])

        for element in elements:
            # Check if element has a label.
            label = element.get("label", {}).get("value")

            if label not in label2idx:
                # Skip this element if it is not in the label2idx mapping.
                continue

            idx = label2idx[label]

            element_type = element["type"]

            if element_type == "polyline":
                points = element.get("points", [])

                if len(points):
                    contours[idx]["points"].append(
                        np.array(points, dtype=int)[:, :2]
                    )

                holes = element.get("holes", [])

                if len(holes):
                    for hole in holes:
                        contours[idx]["holes"].append(
                            np.array(hole, dtype=int)[:, :2]
                        )
            elif element_type == "rectangle":
                # Convert the rectangle coordinates to polygon ones.
                x1, y1 = element["center"][:2]
                x2 = x1 + element["width"]
                y2 = y1 + element["height"]

                contours[idx]["points"].append(
                    np.array(
                        [[x1, y1], [x2, y1], [x2, y2], [x1, y2]], dtype=int
                    )
                )

    # Create directory to save tile images and masks.
    save_dir = Path(save_dir)
    img_dir = save_dir / "images"
    mask_dir = save_dir / "masks"
    img_dir.mkdir(parents=True, exist_ok=True)
    mask_dir.mkdir(exist_ok=True)

    # Get the tile source of WSI.
    ts = large_image.getTileSource(str(wsi_fp))

    # Get the metadata of the WSI.
    metadata = ts.getMetadata()

    wsi_w, wsi_h = metadata["sizeX"], metadata["sizeY"]

    # Get the magnification of the WSI.
    wsi_mag = metadata["magnification"]

    # Calculate the tile size at full resolution (1 if mag is None or the same as scan mag).
    if mag is None:
        scan_tile_size = tile_size
        mag = wsi_mag
        sf = 1
    else:
        sf = mag / wsi_mag  # scan mag -> desired mag
        scan_tile_size = int(tile_size / sf)

    # Get a list of x,y coordinates for the tile top left corners.
    xys = [
        (x, y)
        for x in range(0, wsi_w, scan_tile_size)
        for y in range(0, wsi_h, scan_tile_size)
    ]

    with Pool(nproc) as pool:
        jobs = [
            pool.apply_async(
                _process_tile,
                (
                    ts,
                    prepend_name,
                    xy[0],
                    xy[1],
                    scan_tile_size,
                    tile_size,
                    mag,
                    contours,
                    sf,
                    img_dir,
                    mask_dir,
                ),
            )
            for xy in xys
        ]

        tile_info = []

        for job in tqdm(jobs, desc="Tiling..."):
            tile_info.append(job.get())

        return tile_info



# Allow modules to be imported.
from . import (
    object_detection,
    datasets,
    transforms,
    segformer_semantic_segmentation,
    tiling,
)

# Modules that are imported from * notation.
__all__ = [
    "object_detection",
    "datasets",
    "metrics",
    "transforms",
    "segformer_semantic_segmentation",
    "tiling",
]
